from abc import ABC, abstractmethod
from typing import Any

from ._itypes import GuardFunc, IContext, PostHandle
from .errors import DunglingGuardError

"""
class AuthContext(TypedDict):
    token: str
    user: User

@guard(UserCommand)
class AuthService:
    def validate_user(context: AuthContext, command: UserCommand):
        user = self._get_user(context.token.sub)
        if user.user_id != command.user_id:
            raise ValidationError
        context["user"] = user

async def login(login_form):
    user = authenticate_user(fake_users_db, form_data.username, form_data.password)
    if not user:
        raise HTTPException(status_code=400, detail="Incorrect username or password")
    access_token_expires = timedelta(minutes=ACCESS_TOKEN_EXPIRE_MINUTES)
    access_token = create_access_token(
        data={"sub": user.username, "scopes": form_data.scopes},
        expires_delta=access_token_expires,
    )
    return Token(access_token=access_token, token_type="bearer"

class AuthenticatedCommand:
    token: st

def validate_token(command: AuthenticatedCommand) -> AccessToken:
    try:
        token_dict = decrypt_jwt(command.token)
        return AccessToken.model_validate(token_dict)
    except (JWTError, ValidationError) as e:
        raise InvalidCredentialError(
            "Your access token is invalid, check for expiry"
        ) from 

async def get_current_user(aut_repo: AuthRepo, jwt_token: str) -> User:
    token = validate_token(jwt_token)
    user_id = token.sub
    user = await self._auth_repo.get(user_id)
    if not user:
        raise UserNotFoundError(user_id=user_id)
    return user
"""


# TODO: should be parent class of Guard
class BaseGuard(ABC):
    """
    An abstract class for advanced user-defined guard.

    Example
    ---

    ```py
    from typing import Any, MutableMapping
    from loguru import Logger

    type Context = MutableMapping[str, Any]


    class LogginGuard(BaseGuard):
        _next_guard: GuardFunc

        def __init__(self, next_guard: GuardFunc | None = None, *, logger: Any):
            super().__init__(next_guard)
            self._logger = logger

        async def __call__(self, message: Any, context: Context):
            if (request_id := context.get("request_id")) is None:
                context["request_id"] = request_id = str(uuid4())

            with logger.contextualize(request_id=request_id):
                try:
                    response = await self._next_guard(message, context)
                except Exception as exc:
                    logger.error(exc)
                else:
                    logger.success(f"request: {request_id}, got response `{response}`")
                    return response
    ```
    """

    _next_guard: GuardFunc | None

    def __init__(self, next_guard: GuardFunc | None = None):
        self._next_guard = next_guard

    @property
    def next_guard(self) -> GuardFunc | None:
        return self._next_guard

    def __repr__(self):
        base = f"{self.__class__.__name__}("
        if self._next_guard:
            base += f"next_guard={self._next_guard}"
        base += ")"
        return base

    def chain_next(self, next_guard: GuardFunc, /) -> None:
        self._next_guard = next_guard

    async def __call__(self, command: Any, context: IContext) -> Any:
        """
        subclass of `BaseGuard` should override __call__ similar to following:

        ```py
        async def __call__(self, message: Any, context: IContext):
            # write your pre-handle logic
            response = await super().__call__(message, context)
            # write your post-handle logic
            return response
        ```
        """
        if not self._next_guard:
            raise DunglingGuardError(self)
        return await self._next_guard(command, context)


class Guard(BaseGuard):
    def __init__(
        self,
        next_guard: GuardFunc | None = None,
        /,
        *,
        pre_handle: GuardFunc | None = None,
        post_handle: PostHandle[Any] | None = None,
    ):
        super().__init__(next_guard)
        self.pre_handle = pre_handle
        self.post_handle = post_handle

    async def __call__(self, command: Any, context: IContext) -> Any:
        if self.pre_handle:
            await self.pre_handle(command, context)

        if not self._next_guard:
            raise DunglingGuardError(self)

        response = await self._next_guard(command, context)
        if self.post_handle:
            return await self.post_handle(command, context, response)
        return response



import pytest

from anywise import Anywise, MessageRegistry
from anywise.errors import InvalidHandlerError
from tests.conftest import (
    CreateUser,
    RemoveUser,
    UserCommand,
    UserCreated,
    UserEvent,
    UserNameUpdated,
)

user_message_registry = MessageRegistry(event_base=UserEvent, command_base=UserCommand)


async def react_to_event(event: UserCreated | UserNameUpdated) -> None:
    print(f"handling {event=}")


async def random_func(cmd: str): ...


class UserService:
    def __init__(self, name: str = "test", *, anywise: Anywise):
        self.name = name
        self._aw = anywise

    async def create_user(self, cmd: CreateUser) -> str:
        assert self.name == "test"
        return "hello"

    async def remove_user(self, cmd: RemoveUser) -> str:
        assert self.name == "test"
        return "goodbye"

    def hello(self) -> str:
        return "hello"


def test_message_registry():
    user_message_registry.register_all(react_to_event, UserService)
    assert user_message_registry.event_mapping[UserNameUpdated]
    assert user_message_registry.command_mapping[CreateUser]


def test_message_register_fail():
    with pytest.raises(InvalidHandlerError):
        user_message_registry.register(random_func)



from asyncio import TaskGroup
from typing import Any

from .anywise import EventContext, EventListeners


async def concurrent_publish(
    msg: Any, context: EventContext | None, subscribers: EventListeners
) -> None:
    if not context:
        context = {}
    async with TaskGroup() as tg:
        for sub in subscribers:
            tg.create_task(sub(msg, context))



from dataclasses import dataclass


@dataclass
class UserCommand:
    user_id: str


@dataclass
class CreateUser(UserCommand):
    user_name: str


@dataclass
class RemoveUser(UserCommand):
    user_name: str


@dataclass
class UpdateUser(UserCommand):
    old_name: str
    new_name: str


@dataclass
class UserEvent: ...


@dataclass
class UserCreated(UserEvent):
    user_name: str


@dataclass
class UserNameUpdated(UserEvent):
    changed_name: str



import typing as ty

from loguru import logger

from anywise import Anywise, BaseGuard, IContext, MessageRegistry
from anywise.sink import InMemorySink
from tests.conftest import CreateUser, UpdateUser, UserCommand

user_registry = MessageRegistry(command_base=UserCommand)

from uuid import uuid4


class LogginGuard(BaseGuard):

    def __init__(self):
        super().__init__()
        self._logger = logger

    async def __call__(self, command: UserCommand | CreateUser, context: IContext):
        if (request_id := context.get("request_id")) is None:
            context["request_id"] = request_id = str(uuid4())

        with logger.contextualize(request_id=request_id):
            try:
                response = await super().__call__(command, context)
            except Exception as exc:
                logger.error(exc)
                raise
            else:
                logger.success(
                    f"Logging request: {request_id}, got response `{response}`"
                )
                return response


user_registry.add_guards(LogginGuard)


@user_registry.pre_handle
async def mark(command: UserCommand, context: IContext) -> None:
    if not context.get("processed_by"):
        context["processed_by"] = ["1"]
    else:
        context["processed_by"].append("1")


class TimeContext(ty.TypedDict):
    processed_by: list[str]


@user_registry.pre_handle
async def timer(command: CreateUser, context: IContext) -> None:
    if not context.get("processed_by"):
        context["processed_by"] = ["2"]
    else:
        context["processed_by"].append("2")


@user_registry
async def handler_create(_: CreateUser, context: TimeContext, anywise: Anywise):
    assert context["processed_by"] == ["1", "2"]
    return "created"


@user_registry
async def handler_update(update_user: UpdateUser, context: TimeContext):
    assert context["processed_by"]
    return "updated"


@user_registry.post_handle
async def post[R](create_user: CreateUser, context: IContext, response: R) -> R:
    assert response in ("created", "updated")
    return response


async def test_guard():
    aw = Anywise(user_registry, sink=InMemorySink())
    await aw.send(CreateUser("1", "2"))
    await aw.send(UpdateUser("1", "2", "3"))



import pytest

from anywise._registry import get_funcmetas, get_methodmetas
from anywise.errors import InvalidHandlerError
from tests.conftest import CreateUser, UpdateUser, UserCommand


async def update_user(cmd: UpdateUser | CreateUser) -> str:
    return "ok"


def test_invalid_handler():

    def test(name: str): ...

    with pytest.raises(InvalidHandlerError):
        get_funcmetas(UserCommand, test)


class UserService:
    def __init__(self):
        self._name = "service"

    @property
    def name(self):
        return self._name

    def hello(self):
        return "hello"

    def create_user(self, command: CreateUser): ...


def test_get_funcmetas():
    results = get_funcmetas(msg_base=UserCommand, func=update_user)
    assert len(results) == 2
    assert results[0].handler == results[1].handler == update_user


def test_get_methodmetas():
    metas = get_methodmetas(UserCommand, UserService)
    assert len(metas) == 1
    assert metas[0].handler is UserService.create_user



from anywise import MessageRegistry
from anywise._visitor import gather_types
from tests.conftest import CreateUser, RemoveUser, UpdateUser, UserCommand

user_registry = MessageRegistry(command_base=UserCommand)


def test_gather_commands():
    b = CreateUser | UserCommand
    c = UserCommand

    all_cmds = {UserCommand, CreateUser, UpdateUser, RemoveUser}

    assert gather_types(b) == gather_types(c) == all_cmds



"""
pre-defined event model and event table
"""

from .eventstore import EventStore as EventStore
from .model import Entity as Entity
from .model import Event as Event
from .model import IEvent as IEvent
from .model import NormalizedEvent as NormalizedEvent
from .table import EventTable as EventTable



import inspect
from collections import defaultdict
from dataclasses import dataclass
from functools import partial
from typing import Any, Callable, Unpack, cast, overload

from ididi import DependencyGraph, INode, INodeConfig

from ._itypes import (
    MISSING,
    FuncMeta,
    HandlerMapping,
    IGuard,
    ListenerMapping,
    Maybe,
    MethodMeta,
    Missed,
)
from ._visitor import Target, gather_types
from .errors import (
    HandlerRegisterFailError,
    InvalidHandlerError,
    MessageHandlerNotFoundError,
    NotSupportedHandlerTypeError,
)
from .guard import Guard, GuardFunc, PostHandle

type GuardMapping = defaultdict[type, list[GuardMeta]]


def get_funcmetas(msg_base: type, func: Callable[..., Any]) -> list[FuncMeta[Any]]:
    params = inspect.Signature.from_callable(func).parameters.values()
    if not params:
        raise MessageHandlerNotFoundError(msg_base, func)

    msg, *rest = params
    is_async: bool = inspect.iscoroutinefunction(func)
    is_contexted: bool = bool(rest) and rest[0].name == "context"
    derived_msgtypes = gather_types(msg.annotation)

    for msg_type in derived_msgtypes:
        if not issubclass(msg_type, msg_base):
            raise InvalidHandlerError(msg_base, msg_type, func)

    metas = [
        FuncMeta[Any](
            message_type=t,
            handler=func,
            is_async=is_async,
            is_contexted=is_contexted,
            ignore=tuple(derived_msgtypes),
        )
        for t in derived_msgtypes
    ]
    return metas


def get_methodmetas(msg_base: type, cls: type) -> list[MethodMeta[Any]]:
    cls_members = inspect.getmembers(cls, predicate=inspect.isfunction)
    method_metas: list[MethodMeta[Any]] = []
    for name, func in cls_members:
        if name.startswith("_"):
            continue
        params = inspect.Signature.from_callable(func).parameters.values()
        if len(params) == 1:
            continue

        _, msg, *rest = params  # ignore `self`
        is_async: bool = inspect.iscoroutinefunction(func)
        is_contexted: bool = bool(rest) and rest[0].name == "context"
        derived_msgtypes = gather_types(msg.annotation)

        if not all(issubclass(msg_type, msg_base) for msg_type in derived_msgtypes):
            continue

        metas = [
            MethodMeta[Any](
                message_type=t,
                handler=func,
                is_async=is_async,
                is_contexted=is_contexted,
                ignore=tuple(derived_msgtypes),
                owner_type=cls,
            )
            for t in derived_msgtypes
        ]
        method_metas.extend(metas)

    if not method_metas:
        raise MessageHandlerNotFoundError(msg_base, cls)

    return method_metas


@dataclass(frozen=True, slots=True, kw_only=True)
class GuardMeta:
    guard_target: type
    guard: IGuard | type[IGuard]


# TODO: Guard Registry
class MessageRegistry[C, E]:
    @overload
    def __init__(
        self,
        *,
        command_base: type[C],
        event_base: type[E] = Missed,
        graph: Maybe[DependencyGraph] = MISSING,
    ) -> None: ...

    @overload
    def __init__(
        self,
        *,
        event_base: type[E],
        command_base: type[C] = Missed,
        graph: Maybe[DependencyGraph] = MISSING,
    ) -> None: ...

    def __init__(
        self,
        *,
        command_base: Maybe[type[C]] = MISSING,
        event_base: Maybe[type[E]] = MISSING,
        graph: Maybe[DependencyGraph] = MISSING,
    ):
        self._command_base = command_base
        self._event_base = event_base
        self._graph = graph or DependencyGraph()

        self.command_mapping: HandlerMapping[Any] = {}
        self.event_mapping: ListenerMapping[Any] = {}

        self.guard_mapping: GuardMapping = defaultdict(list)

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}(command_base={self._command_base}, event_base={self._event_base})"

    @overload
    def __call__[T](self, handler: type[T]) -> type[T]: ...

    @overload
    def __call__[**P, R](self, handler: Callable[P, R]) -> Callable[P, R]: ...

    def __call__[
        **P, R
    ](self, handler: type[R] | Callable[P, R]) -> type[R] | Callable[P, R]:
        return self.register(handler)

    @property
    def graph(self) -> DependencyGraph:
        return self._graph

    @overload
    def factory(self, **config: Unpack[INodeConfig]) -> INode[..., Any]: ...

    @overload
    def factory[
        **P, R
    ](self, factory: INode[P, R] | None = None, **config: Unpack[INodeConfig]) -> INode[
        P, R
    ]: ...

    def factory[
        **P, R
    ](self, factory: INode[P, R] | None = None, **config: Unpack[INodeConfig]) -> INode[
        P, R
    ]:
        if factory is None:
            return cast(INode[P, R], partial(self.factory, **config))

        self._graph.node(**config)(factory)
        return factory

    def _register_commandhanlders(self, handler: Target) -> None:
        if not self._command_base:
            return

        if inspect.isfunction(handler):
            metas = get_funcmetas(self._command_base, handler)
        elif inspect.isclass(handler):
            metas = get_methodmetas(self._command_base, handler)
        else:
            raise NotSupportedHandlerTypeError(handler)

        mapping = {meta.message_type: meta for meta in metas}
        self.command_mapping.update(mapping)

    def _register_eventlisteners(self, listener: Target) -> None:
        if not self._event_base:
            return

        if inspect.isfunction(listener):
            metas = get_funcmetas(self._event_base, listener)
        elif inspect.isclass(listener):
            metas = get_methodmetas(self._event_base, listener)
        else:
            raise NotSupportedHandlerTypeError(listener)

        for meta in metas:
            msg_type = meta.message_type
            if msg_type not in self.event_mapping:
                self.event_mapping[msg_type] = [meta]
            else:
                self.event_mapping[msg_type].append(meta)

    @overload
    def register[T](self, handler: type[T]) -> type[T]: ...

    @overload
    def register[**P, R](self, handler: Callable[P, R]) -> Callable[P, R]: ...

    def register(self, handler: Target):
        try:
            self._register_commandhanlders(handler)
        except HandlerRegisterFailError:
            self._register_eventlisteners(handler)
        return handler

    def register_all(
        self,
        *handlers: Callable[..., Any],
        pre_hanldes: list[GuardFunc] | None = None,
        post_handles: list[PostHandle[Any]] | None = None,
    ) -> None:
        for handler in handlers:
            self.register(handler)

        if pre_hanldes:
            for pre_handle in pre_hanldes:
                self.pre_handle(pre_handle)

        if post_handles:
            for post_handle in post_handles:
                self.post_handle(post_handle)

    # TODO? separate guard registry from message registry
    def _extra_guardfunc_annotation(self, func: Callable[..., Any]) -> type:
        if inspect.isclass(func):
            func_params = list(inspect.signature(func.__call__).parameters.values())
            cmd_type = func_params[1].annotation
        elif inspect.isfunction(func):
            func_params = list(inspect.signature(func).parameters.values())
            cmd_type = func_params[0].annotation
        else:
            raise MessageHandlerNotFoundError(self._command_base, func)
        return cmd_type

    def pre_handle(self, func: GuardFunc) -> GuardFunc:
        target = self._extra_guardfunc_annotation(func)
        meta = GuardMeta(guard_target=target, guard=Guard(pre_handle=func))
        self.guard_mapping[meta.guard_target].append(meta)
        return func

    def post_handle[R](self, func: PostHandle[R]) -> PostHandle[R]:
        target = self._extra_guardfunc_annotation(func)
        meta = GuardMeta(guard_target=target, guard=Guard(post_handle=func))
        self.guard_mapping[meta.guard_target].append(meta)
        return func

    def add_guards(self, *guards: IGuard | type[IGuard]) -> None:
        for guard in guards:
            target = self._extra_guardfunc_annotation(guard)
            meta = GuardMeta(guard_target=target, guard=guard)
            self.guard_mapping[target].append(meta)

    # def guard[
    #     **P, R
    # ](self, func_or_cls: IGuard | type[IGuard]) -> IGuard | type[IGuard]:
    #     """
    #     @registry.guard
    #     async def guard_func(command: Command, context: IContext, next: GuardFunc):
    #         # do something before
    #         response = await next(command, context)
    #         # do something after

    #     @registry.guard
    #     class LoggingGuard(BaseGuard):
    #         async def __call__(self, command: Command, context: IContext):
    #             # do something before
    #             response = await self._next_guard
    #             # do something after
    #     """



# AWS Source
# https://docs.aws.amazon.com/lambda/latest/dg/python-handler.html

# # Define some functions to perform the CRUD operations
# import asyncio
# import typing as ty

# from ..anywise import Anywise

# Event = ty.NewType("Event", dict[str, ty.Any])
# Context = ty.NewType("Context", dict[str, ty.Any])

# anywise = Anywise()


# class UserCreated: ...


# def lambda_handler(event: UserCreated, context: Context):
#     """Provide an event that contains the following keys:
#     - operation: one of the operations in the operations dict below
#     - payload: a JSON object containing parameters to pass to the
#       operation being performed
#     """

#     # operation = event["operation"]
#     # payload = event["payload"]

#     asyncio.run(anywise.send(event))


# GRPC Source
# import grpc
# import asyncio

# async def run() -> None:
#     async with grpc.aio.insecure_channel("localhost:50051") as channel:
#         stub = helloworld_pb2_grpc.GreeterStub(channel)
#         response = await stub.SayHello(helloworld_pb2.HelloRequest(name="you"))

# _cleanup_coroutines = []


# class Greeter(helloworld_pb2_grpc.GreeterServicer):
#     async def SayHello(
#         self,
#         request: helloworld_pb2.HelloRequest,
#         context: grpc.aio.ServicerContext,
#     ) -> helloworld_pb2.HelloReply:
#         logging.info("Received request, sleeping for 4 seconds...")
#         await asyncio.sleep(4)
#         logging.info("Sleep completed, responding")
#         return helloworld_pb2.HelloReply(message="Hello, %s!" % request.name)


# async def serve() -> None:
#     server = grpc.aio.server()
#     helloworld_pb2_grpc.add_GreeterServicer_to_server(Greeter(), server)
#     listen_addr = "[::]:50051"
#     server.add_insecure_port(listen_addr)
#     logging.info("Starting server on %s", listen_addr)
#     await server.start()

#     async def server_graceful_shutdown():
#         logging.info("Starting graceful shutdown...")
#         # Shuts down the server with 5 seconds of grace period. During the
#         # grace period, the server won't accept new connections and allow
#         # existing RPCs to continue within the grace period.
#         await server.stop(5)

#     _cleanup_coroutines.append(server_graceful_shutdown())
#     await server.wait_for_termination()


# if __name__ == "__main__":
#     logging.basicConfig(level=logging.INFO)
#     loop = asyncio.get_event_loop()
#     try:
#         loop.run_until_complete(serve())
#     finally:
#         loop.run_until_complete(*_cleanup_coroutines)
#         loop.close()

import typing as ty


class ICommandEnvelope(ty.Protocol):
    """
    Meta data of event
    encryption type, schema, key, and serialization format. etc.

    express the whole process chain of command

    command -> handler -> response | errors
    """

    headers: "CommandMeta"
    command: "ICommand"


class CommandMeta:
    # source: uri, post: users/sessions/chats
    command_source: ...  # commands/command/v1,
    answer: type
    errors: list[Exception]


class ICommand:
    entity_id: str
    command_id: str


class HttpSource: 
    "Uvicorn"
    ...


class KafkaSource: ...


class GRPCSource: ...
"""
class KafkaSource:
    def __init__(self, client, anywise: Anywise)
        self.client = client
        self.anywise = anywise

    
    async def start(self): ...
    async def stop(self): ...

    
    async def polling(self):
        ...

        
async def main():
    async with KafkaSource() as source:
        await source.polling()
"""


class RedisJobQueue:
    "https://arq-docs.helpmanual.io/"

    """
    async def main():

    redis = await create_pool(REDIS_SETTINGS)

    source = RedisJobSource(redis, anywise)
        
    # server code
    for url in ('https://facebook.com', 'https://microsoft.com', 'https://github.com'):
        command = DownloadContent(url)
        await source.send(command)
    
    # client code
    async with source:
        await source.run_forever()
    """
    ...



from typing import Annotated

from fastapi import Depends, Request

from ..anywise import Anywise


class InvalidAppStateError(Exception):
    def __init__(self):
        super().__init__("Make sure `anywise` exist in app.state")


def get_anywise(r: Request) -> Anywise:
    try:
        anywise = r.scope["state"]["anywise"]
    except KeyError:
        raise InvalidAppStateError()
    return anywise


FastWise = Annotated[Anywise, Depends(get_anywise)]


# class FastAPISourceConfig(TypedDict):
#     path: str
#     http_method: Literal["GET", "POST", "PATCH", "DELETE"]

#     # return_type: type
#     # response_class: type


# class FastAPICommand(Protocol):
#     __source_config__: ClassVar[FastAPISourceConfig]


# @dataclass
# class UserCommand: ...


# @dataclass
# class CreateUser(UserCommand):
#     __source_config__ = FastAPISourceConfig(path="/users", http_method="POST")

#     """
#     converts this to
#     router = APIRouter()

#     @router.post("/users")
#     async def create_user(command: CreateUser, anywise: AnyWise):
#         await anywise.send(command)
#     """


# @dataclass
# class UpdateUser(UserCommand):
#     """
#     converts this to
#     router = APIRouter()

#     @router.patch("/users/{entity_id}")
#     async def update_user(command: UpdateUser, anywise: AnyWise):
#         await anywise.send(command)
#     """

#     __source_config__ = FastAPISourceConfig(
#         path="/users/{entity_id}", http_method="PATCH"
#     )

#     entity_id: str  # Field(alias="user_id")


# def autoroute(message: FastAPICommand):
#     """
#     generate routes from registry

#     app = FastAPI()
#     app.include_router(autoroute(UserCommand))
#     app.include_router(autoroute(InventoryCommand))
#     """

#     try:
#         config = message.__source_config__
#     except AttributeError:
#         raise

#     # APIRoute
#     route = router.add_api_route()

#     """
#     for base in type(message).__mro__:
#         merge_config(base.__source_config__)

#     base_mapping: dict[type, Router]
#     for base in type(message).__mro__:
#         ...

#     generate APIRoute based on Message.__source_config__
#     include route in router

#     UserCommand -> APIRouter
#     CreateUser -> APIRoute


#     type CommandGraph = dict[type, list[CommandGraph]]

#     command_graph: CommandGraph = {
#         UserCommand: [
#             {
#                 CreateUser: [
#                     {CreateUserSub: []}
#                 ]
#             }
#         ]
#     }
#     """



import inspect
import sys
from types import UnionType
from typing import Annotated, Any, Callable, Union, get_args, get_origin

from .errors import InvalidMessageTypeError

type Target = type | Callable[..., Any]


# def _extract_from_module(base_msg_type: type, module: ModuleType):
# ...
# def auto_collect(msg_type: type, dir: pathlib.Path):
#     """
#     scan through dir, looking for function / methods
#     that contain subclass of msg_type as param of signature
#     record its return type
#     constrcut a anywise.pyi stub file along the way
#     """


if sys.version_info >= (3, 10):
    UNION_META = (UnionType, Union)
else:
    UNION_META = (Union,)


def all_subclasses(cls: type) -> set[type]:
    return set(cls.__subclasses__()).union(
        *[all_subclasses(c) for c in cls.__subclasses__()]
    )


def gather_types(annotation: Any) -> set[type]:
    """
    Recursively gather all types from a type annotation, handling:
    - Union types (|)
    - Annotated types
    - Direct types
    """
    types: set[type] = set()

    # Handle None case
    if annotation is inspect.Signature.empty:
        return types

    origin = get_origin(annotation)
    if not origin:
        types.add(annotation)
        types |= all_subclasses(annotation)
    else:
        # Union types (including X | Y syntax)
        if origin is Annotated:
            # For Annotated[Type, ...], we only care about the first argument
            param_type = get_args(annotation)[0]
            types.update(gather_types(param_type))
        elif origin in UNION_META:  # Union[X, Y] and X | Y
            for arg in get_args(annotation):
                types.update(gather_types(arg))
        else:
            # Generic type, e.g. List, Dict, etc.
            raise InvalidMessageTypeError(origin)
    return types



import typing as ty

from fastapi import APIRouter, FastAPI
from sqlalchemy.ext.asyncio import AsyncEngine

from anywise import Anywise
from anywise.integration.fastapi import FastWise

from .message import CreateTodo, ListTodoEvents, ListTodos, RenameTodo, Todo
from .table import create_tables
from .todo import registry

todo_router = APIRouter()


@todo_router.get("/todos")
async def read_todos(anywise: FastWise) -> list[Todo]:
    res = await anywise.send(ListTodos())
    return res


@todo_router.get("/events")
async def read_todo_events(todo_id: str, anywise: FastWise) -> list[dict[str, ty.Any]]:
    return await anywise.send(ListTodoEvents(todo_id=todo_id))


@todo_router.post("/todos")
async def _(command: CreateTodo, anywise: FastWise) -> str:
    res = await anywise.send(command)
    return res


@todo_router.put("/todos/{todo_id}")
async def _(command: RenameTodo, anywise: FastWise):
    return await anywise.send(command)


class AppState(ty.TypedDict):
    anywise: Anywise


async def lifespan(app: FastAPI) -> ty.AsyncGenerator[AppState, None]:
    anywise = Anywise()
    anywise.include(registry)
    async with anywise.scope("app") as app_scope:
        app_scope.register_dependent(anywise, Anywise)
        engine = await app_scope.resolve(AsyncEngine)
        await create_tables(engine)
        yield {"anywise": anywise}


def app_factory():
    VERSION = "1"
    root_path = f"/api/v{VERSION}"
    app = FastAPI(lifespan=lifespan, version=VERSION, root_path=root_path)
    app.include_router(todo_router)
    return app



from typing import Any, Callable

from ._itypes import IGuard


class AnyWiseError(Exception): ...


class NotSupportedHandlerTypeError(AnyWiseError):
    def __init__(self, handler: Any):
        super().__init__(f"{handler} of type {type(handler)} is not supported")


class HandlerRegisterFailError(AnyWiseError): ...


class InvalidMessageTypeError(HandlerRegisterFailError):
    def __init__(self, msg_type: type):
        super().__init__(f"{msg_type} is not a valid message type")


class MessageHandlerNotFoundError(HandlerRegisterFailError):
    def __init__(self, base_type: Any, handler: Any):
        super().__init__(f"can't find param of type `{base_type}` in {handler}")


class InvalidHandlerError(HandlerRegisterFailError):
    def __init__(self, basetype: type, msg_type: type, handler: Callable[..., Any]):
        msg = f"{handler} is receiving {msg_type}, which is not a valid subclass of {basetype}"
        super().__init__(msg)


class UnregisteredMessageError(AnyWiseError):
    def __init__(self, msg: Any):
        super().__init__(f"Handler for message {msg} is not found")


class DunglingGuardError(AnyWiseError):
    def __init__(self, guard: IGuard):
        super().__init__(f"Dangling guard {guard}, most likely a bug")



"""
a command line interface for anywise
maybe we can also build a tui for anywise


anywise --help

anywise todo_app.main
# this would serve the message source, such as kafka, rabbitmq, etc


anywise todo_app.main --sink
# this would poll the outbox table and send the messages to the event sink

"""

# from sqlalchemy import select, update, func
# from sqlalchemy.ext.asyncio import AsyncEngine

# class OutboxEventProcessor:
#     def __init__(self, engine: AsyncEngine):
#         self._engine = engine

#     def fetch_pending_events(self):
#         """Fetch all pending events by joining outbox_events with events."""
#         stmt = (
#             select(
#                 outbox_events.c.id,
#                 outbox_events.c.event_id,
#                 outbox_events.c.status,
#                 outbox_events.c.retry_count,
#                 events.c.aggregate_id,
#                 events.c.event_type,
#                 events.c.payload
#             )
#             .join_from(outbox_events, events, outbox_events.c.event_id == events.c.id)
#             .where(outbox_events.c.status == 'pending')
#             .order_by(outbox_events.c.id)  # Optional: order by outbox event ID
#         )
#         async with self._engine.connect() as connection:
#             # Build the SQLAlchemy Core SELECT statement with JOIN
#             result = connection.execute(stmt)
#             events = result.fetchall()

#         return events

#     def process_event(self, event):
#         """Simulate processing an event (e.g., sending it to a message queue)."""

#     def update_event_status(self, event_id, status, retry_count):
#         """Update the status of the event in the outbox_events table."""
#         with self._engine.connect() as connection:
#             stmt = (
#                 update(outbox_events)
#                 .where(outbox_events.c.event_id == event_id)
#                 .values(status=status, retry_count=retry_count + 1, processed_at=func.now())
#             )
#             connection.execute(stmt)
#             print(f"Event {event_id} status updated to {status}.")

#     def poll_outbox():
#         """Periodically poll the outbox_events table for pending events."""
#         while True:
#             print("Polling for pending events...")
#             events_to_process = fetch_pending_events()

#             if events_to_process:
#                 print(f"Found {len(events_to_process)} pending events to process.")
#                 for event in events_to_process:
#                     try:
#                         # Process the event
#                         process_event(event)

#                         # Update status to 'sent' after successful processing
#                         update_event_status(event['event_id'], 'sent', event['retry_count'])
#                     except Exception as e:
#                         print(f"Error processing event {event['event_id']}: {e}")

#                         # If processing fails, increment retry count and set status to 'failed'
#                         retry_count = event['retry_count'] + 1
#                         update_event_status(event['event_id'], 'failed', retry_count)
#             else:
#                 print("No pending events found, waiting for the next poll...")

#             # Wait for a few seconds before polling again (adjust as needed)
#             time.sleep(5)  # Poll every 5 seconds



"""
# RFC9457
ref: https://www.rfc-editor.org/rfc/rfc9457.html

class ErrorResponse[Exc]:
    status: 422
    # The "status" member is a JSON number indicating the HTTP status code

    type: https://example.net/validation-error
    # human-readable documentation for the problem type 

    title: "You do not have enough credit.",
    #  a short, human-readable summary of the problem type.

    detail: "Your current balance is 30, but that costs 50."
    # a human-readable explanation specific to this occurrence of the problem.


    instance: "/account/12345/msgs/abc",
    # a URI reference that identifies the specific occurrence of the problem.
    # When the "instance" URI is dereferenceable, the problem details object can be fetched from it. It might also return information about the problem occurrence in other formats through use of proactive content negotiation

    def __init__(self, exc: Exc):
        error_base = "api/v1/errors/"

        self.type = error_base + keba_case(exc.__class__.__name__)
        self.title = exc.__class__.__doc__
        self.detail = exc.detail
        self.instance = exc.entity_path + exc.entity_id

class Exc(ty.Protocol):
    __doc__: str
    __status__: int

    detail: str
    entity_id: str
"""

import typing as ty


class ErorrResponse[Exc: Exception]:
    status: int
    type: str
    title: str
    detail: str
    instance: str

    def __init__(self, exc: Exc):
        """
        error_base = "api/v1/errors/"

        self.type = error_base + keba_case(exc.__class__.__name__)
        self.title = exc.__class__.__doc__
        self.detail = exc.detail
        self.instance = exc.entity_path + exc.entity_id

        """


type Result[Ok, Err: Exception] = ty.Annotated[Ok, Err]
"""
Carry normal return and exception  

type CreateUserError = (
    UserNotFoundError
    | InvalidStateError
    | ReplicatedUserError
    | BannedUserError
    | LongNameErrror
    | AnyNewError
)

async def create_user(cmd: CreateUser, user_repo) -> Result[str, CreateUserError]: ...

async def create_user(
    cmd: CreateUser,
    user_repo: UserRepo,
    token_registry: TokenRegistry,
    event_store: EventStore,
) -> Result[str, CreateUserError]:
"""


class EUserNotCreated(Exception): ...


class EDuplicatedUser(Exception): ...


type CreateUserError = EUserNotCreated | EDuplicatedUser


async def create_user() -> Result[int, CreateUserError]:
    return 3



from datetime import UTC, datetime
from functools import singledispatchmethod
from typing import Any, ClassVar, Final, Protocol, Self, Sequence, TypedDict, cast
from uuid import uuid4

__EventTypeRegistry__: Final[dict[str, type]] = {}

# TODO: rename to folder message
# add a Command Model, with type registry


def uuid_factory() -> str:
    return str(uuid4())


def utc_now() -> str:
    return datetime.now(UTC).isoformat()


def deafult_typeid(cls: type["Event"]) -> str:
    """
    generate a type_id based on event class

    e.g.
    "demo.domain.event:UserCreated"
    """
    # return cls.__type_id__()
    return f"{cls.__module__}:{cls.__name__}"


def all_subclasses(cls: type) -> set[type]:
    return set(cls.__subclasses__()).union(
        *[all_subclasses(c) for c in cls.__subclasses__()]
    )


def get_event_cls(event_type_id: str) -> type["Event"]:
    try:
        return __EventTypeRegistry__[event_type_id]
    except KeyError:
        __EventTypeRegistry__.update(
            {deafult_typeid(cls): cls for cls in all_subclasses(Event)}
        )
        # if fail again just raise

    try:
        return __EventTypeRegistry__[event_type_id]
    except KeyError:
        raise UnregisteredEventError(event_type_id)


class UnregisteredEventError(Exception):
    def __init__(self, type_id: str):
        name = type_id.replace(":", ".")
        super().__init__(f"event {name} is not registered")


class IEvent(Protocol):
    @property
    def event_id(self) -> str: ...

    @property
    def entity_id(self) -> str: ...

    @property
    def timestamp(self) -> str: ...

    @classmethod
    def __type_id__(cls) -> str: ...

    def __normalized__(self) -> "NormalizedEvent": ...


class NormalizedEvent(TypedDict):
    # classfields
    event_type: str
    source: str
    version: str

    # base fields
    event_id: str
    entity_id: str
    timestamp: str

    # current only fields
    event_body: dict[str, Any]


try:
    from msgspec import Struct
except ImportError:
    pass
else:

    from msgspec import field as msgspec_field

    class Event(Struct, frozen=True, kw_only=True):
        __source__: ClassVar[str] = "unspecified"  # project name, like demo
        __version__: ClassVar[str] = "1"  # specversion

        entity_id: str
        event_id: str = msgspec_field(default_factory=uuid_factory)
        timestamp: str = msgspec_field(default_factory=utc_now)

        @classmethod
        def __type_id__(cls) -> str:
            "generate a unique id for event type, e.g. 'demo.domain.event:UserCreated', if class is renamed make sure to update this method"
            return deafult_typeid(cls)

        def __normalized__(self) -> NormalizedEvent:
            base_fields = Event.__struct_fields__
            current_only_fields = self.__struct_fields__[: -len(base_fields)]

            mapping = {f: getattr(self, f) for f in base_fields}
            event_body = {f: getattr(self, f) for f in current_only_fields}

            mapping["event_type"] = deafult_typeid(self.__class__)
            mapping["version"] = self.__version__
            mapping["source"] = self.__source__
            mapping["event_body"] = event_body
            return cast(NormalizedEvent, mapping)

    class Entity(Struct, kw_only=True):
        entity_id: str

        @singledispatchmethod
        @classmethod
        def apply(cls, event: Event) -> "Self":
            raise NotImplementedError

        @apply.register
        def _(self, _: object) -> "Self":
            raise NotImplementedError

        @classmethod
        def rebuild(cls, events: Sequence[Event]) -> "Self":
            create, rest = events[0], events[1:]
            self = cls.apply(create)

            for e in rest:
                self.apply(e)

            return self


# TODO: pydantic version



import sqlalchemy as sa
from sqlalchemy.ext import asyncio as saio

from anywise.messages.table import TableBase


class TodoTable(TableBase):
    __tablename__: str = "todos"

    todo_id = sa.Column("todo_id", sa.String, unique=True, nullable=False, index=True)
    title = sa.Column("title", sa.String, unique=False, nullable=False)
    content = sa.Column("content", sa.String, unique=False)
    is_completed = sa.Column("is_completed", sa.Boolean, default=False)


async def create_tables(engine: saio.AsyncEngine):
    async with engine.begin() as conn:
        await conn.run_sync(TableBase.metadata.create_all)



from collections import defaultdict
from typing import AsyncGenerator

from sqlalchemy import insert, select
from sqlalchemy.ext.asyncio import AsyncEngine

from .model import IEvent
from .table import EventTable, event_to_mapping, mapping_to_event


class EventStore:
    def __init__(self, engine: AsyncEngine):
        self._engine = engine

    async def add(self, event: IEvent):
        stmt = insert(EventTable).values(**event_to_mapping(event))
        async with self._engine.begin() as conn:
            await conn.execute(stmt)

    async def list_events(self, entity_id: str) -> list[IEvent]:
        stmt = select(EventTable).where(EventTable.entity_id == entity_id)
        async with self._engine.begin() as conn:
            cursor = await conn.execute(stmt)
            mappings = cursor.mappings().all()
            return [mapping_to_event(mapping) for mapping in mappings]

    async def list_all_events(self) -> AsyncGenerator[IEvent, None]:
        stmt = select(EventTable)
        async with self._engine.begin() as conn:
            cursor = await conn.stream(stmt)
            async for row in cursor.mappings():
                yield mapping_to_event(row)

    async def all_event_streams(self) -> AsyncGenerator[list[IEvent], None]:
        grouped = defaultdict[str, list[IEvent]](list)
        current_id: str = ""
        async for e in self.list_all_events():
            if current_id and current_id != e.entity_id:
                yield grouped[current_id]
                del grouped[current_id]  # Free memory for the yielded group
            current_id = e.entity_id
            grouped[e.entity_id].append(e)

        # Yield the final group after the loop
        if current_id:
            yield grouped[current_id]

    async def event_stream(
        self, entity_id: str, version: str = "1"
    ) -> list[IEvent] | None:
        stmt = select(EventTable).where(
            EventTable.entity_id == entity_id and EventTable.version == version
        )
        async with self._engine.begin() as conn:
            cursor = await conn.execute(stmt)
            mapping = cursor.mappings().all()
            if not mapping:
                return None
            return [mapping_to_event(row) for row in mapping]



from asyncio import to_thread
from collections import defaultdict
from functools import partial
from types import MappingProxyType, MethodType
from typing import Any, Awaitable, Callable, cast
from weakref import ref

from ididi import AsyncScope, DependencyGraph

from ._itypes import (
    CommandHandler,
    EventContext,
    EventListeners,
    FuncMeta,
    IContext,
    IGuard,
    MethodMeta,
    PublishStrategy,
    Registee,
    SendStrategy,
)
from ._registry import (
    GuardMapping,
    GuardMeta,
    HandlerMapping,
    ListenerMapping,
    MessageRegistry,
)
from ._visitor import gather_types
from .errors import UnregisteredMessageError
from .messages import IEvent
from .sink import IEventSink

# from warnings import warn


async def default_send(
    message: Any, context: IContext | None, handler: Callable[[Any, IContext], Any]
) -> Any:
    if context is None:
        context = dict()
    return await handler(message, context)


async def default_publish(
    message: Any,
    context: EventContext | None,
    listeners: list[Callable[[Any, EventContext], Awaitable[None]]],
) -> None:
    if context is None:
        context = MappingProxyType({})

    for listener in listeners:
        await listener(message, context)


def context_wrapper(origin: Callable[[Any], Any]):
    async def inner(message: Any, _: Any):
        return await origin(message)

    return inner


class InjectMixin:
    def __init__(self, dg: DependencyGraph):
        self._dg = dg

    async def _resolve_meta(self, meta: "FuncMeta[Any]", *, scope: AsyncScope):
        handler = meta.handler

        if not meta.is_async:
            handler = partial(to_thread, cast(Any, handler))

        if isinstance(meta, MethodMeta):
            instance = await scope.resolve(meta.owner_type)
            handler = MethodType(handler, instance)
        else:
            ignore = ("context",) + meta.ignore
            handler = self._dg.entry(ignore=ignore)(handler)

        if not meta.is_contexted:
            handler = context_wrapper(handler)
        return handler


class HandlerManager(InjectMixin):
    def __init__(self, dg: DependencyGraph):
        super().__init__(dg)
        self._handler_metas: dict[type, FuncMeta[Any]] = {}
        self._guard_mapping: GuardMapping = defaultdict(list)
        self._global_guard: list[GuardMeta] = []

    def include_handlers(self, command_mapping: HandlerMapping[Any]):
        handler_mapping = {msg_type: meta for msg_type, meta in command_mapping.items()}
        self._handler_metas.update(handler_mapping)

    def include_guards(self, guard_mapping: GuardMapping):
        # gather commands
        for origin_target, guard_meta in guard_mapping.items():
            if origin_target is Any or origin_target is object:
                self._global_guard.extend(guard_meta)
            else:
                # TODO: move it to guard registry
                drived_targets = gather_types(origin_target)
                for target in drived_targets:
                    self._guard_mapping[target].extend(guard_meta)

    async def _chain_guards(
        self,
        msg_type: Any,
        handler: Callable[..., Any],
        *,
        scope: AsyncScope,
    ) -> CommandHandler:
        command_guards = self._global_guard + self._guard_mapping[msg_type]

        if not command_guards:
            return handler

        guards: list[IGuard] = [
            (
                await scope.resolve(meta.guard)
                if isinstance(meta.guard, type)
                else meta.guard
            )
            for meta in command_guards
        ]

        for i in range(len(guards) - 1):
            guards[i].chain_next(guards[i + 1])

        guards[-1].chain_next(handler)
        return guards[0]

    def get_handler(self, msg_type: type) -> CommandHandler | None:
        try:
            meta = self._handler_metas[msg_type]
        except KeyError:
            return None
        else:
            return meta.handler

    async def resolve_handler(self, msg_type: type, scope: AsyncScope):
        try:
            meta = self._handler_metas[msg_type]
        except KeyError:
            raise UnregisteredMessageError(msg_type)

        resolved_handler = await self._resolve_meta(meta, scope=scope)
        guarded_handler = await self._chain_guards(
            msg_type, resolved_handler, scope=scope
        )
        return guarded_handler


class ListenerManager(InjectMixin):
    def __init__(self, dg: DependencyGraph):
        super().__init__(dg)
        self._listener_metas: dict[type, list[FuncMeta[Any]]] = dict()

    def include_listeners(self, event_mapping: ListenerMapping[Any]):
        listener_mapping = {
            msg_type: [meta for meta in metas]
            for msg_type, metas in event_mapping.items()
        }
        for msg_type, metas in listener_mapping.items():
            if msg_type not in self._listener_metas:
                self._listener_metas[msg_type] = metas
            else:
                self._listener_metas[msg_type].extend(metas)

    def get_listeners(self, msg_type: type) -> EventListeners:
        try:
            listener_metas = self._listener_metas[msg_type]
        except KeyError:
            return []
        else:
            return [meta.handler for meta in listener_metas]

    async def resolve_listeners(
        self, msg_type: type, *, scope: AsyncScope
    ) -> EventListeners:
        try:
            listener_metas = self._listener_metas[msg_type]
        except KeyError:
            raise UnregisteredMessageError(msg_type)
        else:
            resolved_listeners = [
                await self._resolve_meta(meta, scope=scope) for meta in listener_metas
            ]
            return resolved_listeners


class Inspect:
    """
    a util class for inspecting anywise
    """

    def __init__(self, aw: "Anywise"):
        self._aw = ref(aw)

    def __getitem__(self, key: type) -> CommandHandler | EventListeners | None:
        aw = self._aw()

        if aw is None:
            raise Exception(f"{self.__class__.__name__} should not outlive anywise")

        if handler := aw.handler_manager.get_handler(key):
            return handler

        if listeners := aw.listener_manager.get_listeners(key):
            return listeners


class Anywise(InjectMixin):
    """

    - dependency_graph: `DependencyGraph`

    - send_strategy: `Callable[[Any, MutableMapping[Any, Any], CommandHandler], Any]`

        ```py
        async def sender(msg: Any, context: CommandContext, handler: CommandHandler) -> Any:
            await handler(msg, context)
        ```

    - publish_strategy: `Callable[[Any, Mapping[Any, Any], EventListeners], Awaitable[None]]`

        ```py
        async def publisher(msg: Any, context: Mapping[Any, Any], listeners: EventListeners)->None:
            for listener in listeners:
                await listener(msg, context)
        ```

    """

    _sender: SendStrategy
    _publisher: PublishStrategy

    def __init__(
        self,
        *registries: MessageRegistry[Any, Any],
        dependency_graph: DependencyGraph | None = None,
        sink: IEventSink[IEvent] | None = None,
        sender: SendStrategy = default_send,
        publisher: PublishStrategy = default_publish,
    ):
        dependency_graph = dependency_graph or DependencyGraph()
        super().__init__(dependency_graph)
        self.handler_manager = HandlerManager(self._dg)
        self.listener_manager = ListenerManager(self._dg)

        self._sender = sender
        self._publisher = publisher
        self._sink = sink
        self._dg.register_singleton(self)

        self.include(*registries)

    # async def __enter__(self):
    #     """create an global scope and create resource"""
    #     # scope = self._dg.scope("anywise")

    # async def __aexit__(
    #     self,
    #     exc_type: type[Exception] | None,
    #     exc: Exception | None,
    #     exc_tb: Any | None,
    # ): ...

    def register(
        self, message_type: type | None = None, registee: Registee | None = None
    ) -> None:
        """
        register a function, a class, a module, or a package.

        anywise.register(create_user)
        anywise.register(UserCommand, UserService)
        anywise.register(UserCommand, user_service) # module / package

        NOTE: a package is a module with __path__ attribute
        """

    @property
    def inspect(self) -> Inspect:
        return Inspect(self)

    def register_singleton(self, singleton: Any):
        self._dg.register_singleton(singleton)

    def include(self, *registries: MessageRegistry[Any, Any]) -> None:
        for msg_registry in registries:
            self._dg.merge(msg_registry.graph)
            self.handler_manager.include_handlers(msg_registry.command_mapping)
            self.handler_manager.include_guards(msg_registry.guard_mapping)
            self.listener_manager.include_listeners(msg_registry.event_mapping)
        self._dg.static_resolve_all()

    def scope(self, name: str | None = None):
        return self._dg.scope(name)

    async def send(self, msg: object, *, context: IContext | None = None) -> Any:
        # TODO: msg, context, scope
        scope_proxy = self._dg.use_scope("message", create_on_miss=True, as_async=True)

        async with scope_proxy as scope:
            handler = await self.handler_manager.resolve_handler(type(msg), scope)
            return await self._sender(msg, context, handler)

    async def publish(
        self, msg: object, *, context: EventContext | None = None
    ) -> None:
        scope_proxy = self._dg.use_scope("message", create_on_miss=True, as_async=True)

        async with scope_proxy as scope:
            resolved_listeners = await self.listener_manager.resolve_listeners(
                type(msg), scope=scope
            )
            return await self._publisher(msg, context, resolved_listeners)

    async def sink(self, event: Any):
        if self._sink is None:
            raise Exception("sink is not set")
        await self._sink.sink(event)


"""

def source_factory():
    anywise = Anywise()
    source = KafkaSource(
        anywise=anywise
    )


```bash
anywise main.source_factory
```
"""



VERSION = "0.1.7"


from ididi import INodeConfig as INodeConfig
from ididi import use as use

from ._itypes import GuardFunc as GuardFunc
from ._itypes import IContext as IContext
from ._itypes import IGuard as IGuard
from ._registry import MessageRegistry as MessageRegistry
from .anywise import Anywise as Anywise
from .anywise import CommandHandler as CommandHandler
from .anywise import EventListeners as EventListeners
from .anywise import PublishStrategy as PublishStrategy
from .anywise import SendStrategy as SendStrategy
from .guard import BaseGuard as BaseGuard
from .strategies import concurrent_publish as concurrent_publish

# from . import integration as integration



from typing import Any, Mapping

import sqlalchemy as sa
from sqlalchemy import orm as sa_orm
from sqlalchemy.ext import asyncio as saio
from sqlalchemy.sql import func

from .model import IEvent, NormalizedEvent, get_event_cls

TABLE_RESERVED_VARS: set[str] = {
    "id",  # primary key
    "version",
    "source",
    "event_type",
    "event_body",
    "gmt_created",
    "gmt_modified",
}
"Values that exist in event table but should be ignored to rebuild the event model."


def declarative(cls: type) -> type[sa_orm.DeclarativeBase]:
    """
    A more pythonic way to declare a sqlalchemy table
    """

    return sa_orm.declarative_base(cls=cls)


@declarative
class TableBase:
    "Exert constraints on table creation, and reduce duplicate code"
    id = sa.Column("id", sa.Integer, primary_key=True, autoincrement=True)
    gmt_modified = sa.Column(
        "gmt_modified", sa.DateTime, server_default=func.now(), onupdate=func.now()
    )
    gmt_created = sa.Column("gmt_created", sa.DateTime, server_default=func.now())


class EventTable(TableBase):
    __tablename__: str = "events"
    __table_args__: tuple[Any] = (
        sa.Index("idx_events_entity_id_version", "entity_id", "version"),
    )

    event_id = sa.Column(
        "event_id", sa.String, index=False, nullable=False, unique=True
    )  # entity_id of the aggregate root
    event_type = sa.Column("event_type", sa.String)
    event_body = sa.Column("event_body", sa.JSON)
    source = sa.Column("source", sa.String, nullable=False)
    entity_id = sa.Column("entity_id", sa.String, index=True, nullable=False)
    timestamp = sa.Column("timestamp", sa.String)
    version = sa.Column("version", sa.String)
    # version of the aggregate root entity


# class OutBoxEvents(TableBase):
#     """
#     id BIGINT PRIMARY KEY AUTO_INCREMENT,       -- Unique outbox record ID
#     event_id BIGINT NOT NULL,                   -- Foreign key to the events table
#     status ENUM('pending', 'sent', 'failed') DEFAULT 'pending',  -- Delivery status
#     processed_at TIMESTAMP NULL,                -- Timestamp when the event was processed or sent
#     retry_count INT DEFAULT 0,                  -- Number of retries
#     created_at TIMESTAMP DEFAULT CURRENT_TIMESTAMP, -- When the outbox record was created
#     CONSTRAINT fk_event_id FOREIGN KEY (event_id) REFERENCES events(id) -- Link to the event table
#     """

#     __tablename__: str = "outbox_events"

#     event_id = sa.Column(
#         "event_id", sa.String, index=False, nullable=False, unique=True
#     )
#     status = sa.Column(sa.Enum("pending", "sent", "failed"), default="pending")
#     processed_at = sa.Column("processed_at", sa.DateTime, nullable=True)
#     retry = sa.Column("retry", sa.Integer, default=0)


async def create_tables(engine: saio.AsyncEngine):
    async with engine.begin() as conn:
        await conn.run_sync(TableBase.metadata.create_all)


def event_to_mapping(event: IEvent) -> NormalizedEvent:
    return event.__normalized__()


def mapping_to_event(row_mapping: Mapping[Any, Any]) -> IEvent:
    type_id = row_mapping["event_type"]
    event_cls = get_event_cls(type_id)
    mapping = {k: v for k, v in row_mapping.items() if k not in TABLE_RESERVED_VARS}
    body = row_mapping["event_body"]
    return event_cls(**mapping, **body)



from asyncio.queues import Queue
from typing import Protocol, Sequence

from ..messages import IEvent

# class AbstractSink:
#     def sink(self, event: IEvent) -> None:
#         raise NotImplementedError

# class EventSink:
#     """
#     anywise = Anywise(sink=FileSink())

#     """

#     def sink(self, event: IEvent) -> None: ...


# class FileSink(EventSink):
#     "send event to a file"
#     ...


# class WebSink(EventSink):
#     "send event to a remote http service"
#     ...


# class KafkaSink(EventSink):
#     "send event to kafka"


# class DBSink(EventSink):
#     "save events to database"


class IEventSink[EventType](Protocol):

    async def sink(self, event: EventType | Sequence[EventType]):
        """
        sink an event or a sequence of events to corresponding event sink
        """

    # async def encode(self, event: IEvent) -> str:
    #     ...



class InMemorySink[EventType](IEventSink[EventType]):
    def __init__(self, volume: int = 100):
        self._queue = Queue[IEvent](volume)

    async def sink(self, event: IEvent | Sequence[IEvent]):
        print(f"logging {event}")
        if isinstance(event, Sequence):
            for e in event:
                await self._queue.put(e)
        else:
            await self._queue.put(event)



import typing as ty

from sqlalchemy.ext.asyncio import AsyncConnection, AsyncEngine, create_async_engine
from sqlalchemy.sql import insert, select, update

from anywise import Anywise, MessageRegistry, use
from anywise.messages import EventStore, NormalizedEvent

from .message import (
    CreateTodo,
    ListTodoEvents,
    ListTodos,
    RenameTodo,
    Todo,
    TodoCommand,
    TodoCreated,
    TodoEvent,
    TodoRetitled,
)
from .table import TodoTable

# App layer

registry = MessageRegistry(command_base=TodoCommand, event_base=TodoEvent)

from loguru import logger


def engine_factory() -> AsyncEngine:
    url = "sqlite+aiosqlite:///demo/db.db"
    logger.success("engine created")
    return create_async_engine(url)


async def trans_conn(engine: AsyncEngine) -> ty.AsyncGenerator[AsyncConnection, None]:
    logger.success("conn created")
    async with engine.begin() as conn:
        yield conn


class TodoRepository:
    def __init__(self, engine: AsyncEngine = use(engine_factory)):
        self._engine = engine

    async def add(self, todo: Todo):
        stmt = insert(TodoTable).values(
            todo_id=todo.todo_id,
            title=todo.title,
            content=todo.content,
            is_completed=todo.is_completed,
        )

        async with self._engine.begin() as conn:
            await conn.execute(stmt)

    async def get(self, todo_id: str) -> Todo | None:
        stmt = select(TodoTable).where(TodoTable.todo_id == todo_id)
        async with self._engine.begin() as conn:
            cursor = await conn.execute(stmt)
            mapping = cursor.mappings().one_or_none()

        if mapping:
            return Todo(
                todo_id=mapping["todo_id"],
                title=mapping["title"],
                content=mapping["content"],
                is_completed=mapping["is_completed"],
            )
        return None

    async def retitle(self, todo_id: str, title: str) -> None:
        stmt = update(TodoTable).values(title=title).where(TodoTable.todo_id == todo_id)
        async with self._engine.begin() as conn:
            await conn.execute(stmt)


@registry
async def list_events(query: ListTodoEvents, es: EventStore) -> list[NormalizedEvent]:
    events = await es.list_events(query.todo_id)
    return [e.__normalized__() for e in events]


@registry
class TodoService:
    def __init__(
        self,
        anywise: Anywise,
        repo: TodoRepository,
        es: EventStore,
    ):
        self._repo = repo
        self._es = es
        self._aw = anywise

    async def add_new_todo(self, command: CreateTodo):
        event = TodoCreated(
            entity_id=command.id, title=command.title, content=command.content
        )
        todo = Todo.apply(event)
        await self._repo.add(todo)
        await self._es.add(event)
        await self._aw.publish(event)
        return command.id

    async def rename_todo(self, command: RenameTodo):
        event = TodoRetitled(title=command.title, entity_id=command.todo_id)
        todo = await self._repo.get(command.todo_id)
        if not todo:
            raise KeyError(f"todo with {command.todo_id=} not found")
        todo.apply(event)

        # TODO: make these three a transaction
        await self._repo.retitle(todo.todo_id, todo.title)
        await self._es.add(event)
        await self._aw.publish(event)

    async def list_todos(self, _: ListTodos) -> list[Todo]:
        todos: list[Todo] = []
        async for stream in self._es.all_event_streams():
            todos.append(Todo.rebuild(stream))
        return todos



from dataclasses import dataclass, field
from functools import singledispatchmethod
from typing import Self, Sequence
from uuid import uuid4

from anywise.messages import Event, IEvent


# App Layer
def uuid_factory() -> str:
    return str(uuid4())


@dataclass
class TodoCommand: ...


@dataclass
class ListTodoEvents(TodoCommand):
    todo_id: str


class TodoEvent(Event): ...


@dataclass(kw_only=True)
class CreateTodo(TodoCommand):
    id: str = field(default_factory=uuid_factory)
    title: str
    content: str


@dataclass
class RenameTodo(TodoCommand):
    todo_id: str
    title: str


@dataclass
class ListTodos(TodoCommand): ...


class TodoCreated(TodoEvent):
    title: str
    content: str


class TodoRetitled(TodoEvent):
    title: str


@dataclass
class Todo:
    todo_id: str  # uuid
    title: str
    content: str
    is_completed: bool = False

    @classmethod
    def rebuild(cls, events: Sequence[IEvent]) -> "Todo":
        create, rest = events[0], events[1:]
        self = cls.apply(create)

        for e in rest:
            self.apply(e)

        return self

    @singledispatchmethod
    @classmethod
    def apply(cls, event: TodoCreated) -> "Self":
        return cls(
            todo_id=event.entity_id,
            title=event.title,
            content=event.content,
        )

    @apply.register
    def _(self, event: TodoRetitled) -> "Self":
        self.title = event.title
        return self



import pytest

from anywise import Anywise, MessageRegistry, concurrent_publish, use
from tests.conftest import (
    CreateUser,
    RemoveUser,
    UpdateUser,
    UserCommand,
    UserCreated,
    UserEvent,
    UserNameUpdated,
)

user_message_registry = MessageRegistry(event_base=UserEvent, command_base=UserCommand)


@user_message_registry
class UserService:
    def __init__(self, name: str = "test", *, anywise: Anywise):
        self.name = name
        self._aw = anywise

    async def create_user(self, cmd: CreateUser) -> str:
        assert self.name == "test"
        return "hello"

    async def remove_user(self, cmd: RemoveUser) -> str:
        assert self.name == "test"
        return "goodbye"

    def hello(self) -> str:
        return "hello"


def user_service_factory(asynwise: Anywise) -> "UserService":
    return UserService(name="test", anywise=asynwise)


@user_message_registry
async def update_user(
    cmd: UpdateUser,
    anywise: Anywise,
    service: UserService = use(user_service_factory),
) -> str:
    assert service.hello() == "hello"
    await anywise.publish(UserNameUpdated(cmd.new_name))
    return "ok"


@user_message_registry
async def react_to_event(
    event: UserCreated | UserNameUpdated,
    service: UserService = use(user_service_factory),
) -> None:
    print(f"handling {event=}")


@pytest.fixture(scope="module")
def asynwise() -> Anywise:
    aw = Anywise(user_message_registry, publisher=concurrent_publish)
    return aw


async def test_send_to_method(asynwise: Anywise):
    cmd = CreateUser("1", "user")
    res = await asynwise.send(cmd)
    assert res == "hello"

    rm_cmd = RemoveUser("1", "user")
    res = await asynwise.send(rm_cmd)
    assert res == "goodbye"


async def test_send_to_function(asynwise: Anywise):
    cmd = UpdateUser("1", "user", "new")
    res = await asynwise.send(cmd)
    assert res == "ok"


async def test_event_handler(asynwise: Anywise):
    event = UserCreated("new_name")
    await asynwise.publish(event)



"""
NOTE: unpack might be great for query

"""

import typing as ty


class SQLQuery[Q]: ...


class ProductSalesByWeek:
    """
    SELECT
        sku, quantity, price
    FROM
        products
    WHERE
        sku = :sku
    """

    sku: str


class SalesReport:  # BaseModel
    sku: str
    total_sales: int
    total_prices: int
    avg_sales_daily: float

    # @computed_field
    # def avg_price(self):
    #     return self.total_prices / self.total_sales


class Connection: ...


def query(t: type) -> ty.Callable[..., ty.Any]: ...


async def get_sales_report(
    conn: Connection, query: SQLQuery[ProductSalesByWeek]
) -> SalesReport:
    """
    This function could be auto generated, e.g.

    query_func = generate_query(conn=inject(get_conn), query=SQLQuery[SalesReport])
    """
    ...
    # return await conn.execute(query)



type CommandGraph = dict[type, list[CommandGraph]]


class GraphBuilder:
    def __init__(self):
        self.graph: CommandGraph = {}

    def build_graph(self, command: type) -> CommandGraph:
        """Recursively build a graph of command classes and their subclasses."""

        # TODO: currently only work top to bottom, e.g.: build_graph(UserCommand)
        # make it work bottom to top as well, e.g.: build_graph(CreateUserSub)

        subgraph: CommandGraph = {command: []}

        # Get all direct subclasses
        subclasses = [cls for cls in command.__subclasses__()]

        # For each subclass, build its subgraph and add to parent's children
        for subclass in subclasses:
            subclass_graph = self.build_graph(subclass)
            subgraph[command].append(subclass_graph)

        return subgraph



"Interface, types, type alias, and related stuff"

from collections import defaultdict
from dataclasses import dataclass
from types import ModuleType
from typing import (
    Annotated,
    Any,
    AsyncGenerator,
    Awaitable,
    Callable,
    Final,
    Literal,
    Mapping,
    MutableMapping,
    Protocol,
    TypeGuard,
)

type HandlerMapping[Command] = dict[type[Command], "FuncMeta[Command]"]
type ListenerMapping[Event] = dict[type[Event], list[FuncMeta[Event]]]
type GuardMapping[Command] = defaultdict[type[Command], list[IGuard]]

type IContext = MutableMapping[Any, Any]
type GuardFunc = Callable[[Any, IContext], Awaitable[Any]]
type PostHandle[R] = Callable[[Any, IContext, R], Awaitable[R]]
type EventContext = Mapping[Any, Any]


type CommandHandler = Callable[[Any, IContext], Any] | IGuard
type EventListener = Callable[[Any, EventContext], Any]
type EventListeners = list[EventListener]
type SendStrategy = Callable[[Any, IContext | None, CommandHandler], Any]
type PublishStrategy = Callable[
    [Any, EventContext | None, EventListeners], Awaitable[None]
]

type LifeSpan = Callable[..., AsyncGenerator[Any, None]]


type AnnContext[M] = Annotated[M, "__context__"]


class IPackage(Protocol):
    def __path__(self) -> list[str]: ...


type Registee = IPackage | ModuleType | type | CommandHandler | EventListener


class IGuard(Protocol):

    @property
    def next_guard(self) -> GuardFunc | None: ...

    def chain_next(self, next_guard: GuardFunc, /) -> None:
        """
        self._next_guard = next_guard
        """

    async def __call__(self, command: Any, context: IContext) -> Any: ...


@dataclass(frozen=True, slots=True, kw_only=True)
class FuncMeta[Message]:
    """
    is_async: bool
    is_contexted:
    whether the handler receives a context param
    """

    message_type: type[Message]
    handler: Callable[..., Any]
    is_async: bool
    is_contexted: bool
    ignore: tuple[str | type, ...]


@dataclass(frozen=True, slots=True, kw_only=True)
class MethodMeta[Message](FuncMeta[Message]):
    owner_type: type


type Result[R, E] = Annotated[R, E]
"""
A helper type alias to represent a function that can return either a value or an error.

Example:
---
```py
class UserNotFoundError(Exception): ...


class InvalidStateError(Exception): ...


type SignupError = UserNotFoundError | InvalidStateError


def signup_user(command: CreateUser) -> Result[User, SignupError]:
    return user
```
"""


class _Missed:

    def __str__(self) -> str:
        return "MISSING"

    def __bool__(self) -> Literal[False]:
        return False


Missed: Final[type[_Missed]] = _Missed
MISSING = _Missed()


type Maybe[T] = T | _Missed


def is_provided[T](obj: Maybe[T]) -> TypeGuard[T]:
    return obj is not MISSING
